import {Complex} from './type.js'
import type {
   Dependencies, Signature, Returns, RealType, AliasOf
} from '../interfaces/type.js'

declare module "../interfaces/type" {
   interface Signatures<T> {
      addReal: AliasOf<'add', (a: T, b: RealType<T>) => T>
      divideReal: AliasOf<'divide', (a: T, b: RealType<T>) => T>
   }
}

export const add =
   <T>(dep: Dependencies<'add' | 'complex', T>): Signature<'add', Complex<T>> =>
   (w, z) => dep.complex(dep.add(w.re, z.re), dep.add(w.im, z.im))

export const addReal =
   <T>(dep: Dependencies<'addReal' | 'complex', T>):
      Signature<'addReal', Complex<T>> =>
   (z, r) => dep.complex(dep.addReal(z.re, r), z.im)

export const unaryMinus =
   <T>(dep: Dependencies<'unaryMinus' | 'complex', T>):
      Signature<'unaryMinus', Complex<T>> =>
   z => dep.complex(dep.unaryMinus(z.re), dep.unaryMinus(z.im))

export const conj =
   <T>(dep: Dependencies<'unaryMinus' | 'conj' | 'complex', T>):
      Signature<'conj', Complex<T>> =>
   z => dep.complex(dep.conj(z.re), dep.unaryMinus(z.im))

export const subtract =
   <T>(dep: Dependencies<'subtract' | 'complex', T>):
      Signature<'subtract', Complex<T>> =>
   (w, z) => dep.complex(dep.subtract(w.re, z.re), dep.subtract(w.im, z.im))

export const multiply =
   <T>(dep: Dependencies<
       'add' | 'subtract' | 'multiply' | 'conj' | 'complex', T>):
      Signature<'multiply', Complex<T>> =>
   (w, z) => {
      const mult = dep.multiply
      const realpart = dep.subtract(
         mult(         w.re,  z.re), mult(dep.conj(w.im), z.im))
      const imagpart = dep.add(
         mult(dep.conj(w.re), z.im), mult(         w.im,  z.re))
      return dep.complex(realpart, imagpart)
   }

export const absquare =
   <T>(dep: Dependencies<'absquare', T>
       & Dependencies<'add', Returns<'absquare', T>>):
      Signature<'absquare', Complex<T>> =>
   z => dep.add(dep.absquare(z.re), dep.absquare(z.im))

export const divideReal =
   <T>(dep: Dependencies<'divideReal' | 'complex', T>):
      Signature<'divideReal', Complex<T>> =>
   (z, r) => dep.complex(dep.divideReal(z.re, r), dep.divideReal(z.im, r))

export const reciprocal =
   <T>(dep: Dependencies<'conj' | 'absquare' | 'divideReal', Complex<T>>):
      Signature<'reciprocal', Complex<T>> =>
   z => dep.divideReal(dep.conj(z), dep.absquare(z))

export const divide =
   <T>(dep: Dependencies<'multiply' | 'reciprocal', Complex<T>>):
      Signature<'divide', Complex<T>> =>
   (w, z) => dep.multiply(w, dep.reciprocal(z))

// The dependencies are slightly tricky here, because there are three types
// involved: Complex<T>, T, and RealType<T>, all of which might be different,
// and we have to get it straight which operations we need on each type, and
// in fact, we need `addReal` on both T and Complex<T>, hence the dependency
// with a custom name, not generated via Dependencies<...>
export const sqrt =
   <T>(dep: Dependencies<'equal' | 'conservativeSqrt' | 'unaryMinus', RealType<T>>
       & Dependencies<'zero' | 'complex', T>
       & Dependencies<'absquare' | 're' | 'divideReal', Complex<T>>
       & {
         addTR: Signature<'addReal', T>, 
         addRR: Signature<'add', RealType<T>>,
         addCR: Signature<'addReal', Complex<T>>
      }):
      Signature<'sqrt', Complex<T>> =>
   z => {
      const myabs = dep.conservativeSqrt(dep.absquare(z))
      const r = dep.re(z)
      const negr = dep.unaryMinus(r)
      if (dep.equal(myabs, negr)) {
         // pure imaginary square root; z.im already zero
         return dep.complex(
            dep.zero(z.re), dep.addTR(z.im, dep.conservativeSqrt(negr)))
      }
      const num = dep.addCR(z, myabs)
      const denomsq = dep.addRR(dep.addRR(myabs, myabs), dep.addRR(r, r))
      const denom = dep.conservativeSqrt(denomsq)
      return dep.divideReal(num, denom)
   }

export const conservativeSqrt = sqrt
